@@ -2,23 +2,25 @@ import json
 import os
 import pathlib
 import time
-import warnings
-from typing import Optional, List, Set, Dict, Union
+from typing import Optional, List, Set, Dict, Union, TYPE_CHECKING
 
 import git
 
 from labml import logger, monit
 from labml.internal.configs.base import Configs
-from labml.internal.configs.processor import ConfigProcessor
-from labml.internal.configs.processor_dict import ConfigProcessorDict
+from labml.internal.configs.processor import ConfigProcessor, FileConfigsSaver
 from labml.internal.experiment.experiment_run import Run, struct_time_to_time, struct_time_to_date
 from labml.internal.experiment.watcher import ExperimentWatcher
 from labml.internal.lab import lab_singleton
-from labml.internal.logger import logger_singleton as logger_internal
+from labml.internal.tracker import tracker_singleton as tracker
+from labml.internal.monitor import monitor_singleton as monitor
 from labml.internal.util import is_ipynb, is_colab, is_kaggle
 from labml.logger import Text
 from labml.utils import get_caller_file
+from labml.utils.notice import labml_notice
 
+if TYPE_CHECKING:
+    from labml.internal.tracker.writers.web_api import Writer as WebApiWriter
 
 class ModelSaver:
     def save(self, checkpoint_path: pathlib.Path) -> any:
@@ -34,11 +36,16 @@ class CheckpointSaver:
     def __init__(self, path: pathlib.PurePath):
         self.path = path
         self.model_savers = {}
+        self.__no_savers_warned = False
 
     def add_savers(self, models: Dict[str, ModelSaver]):
         """
         ## Set variable for saving and loading
         """
+        if experiment_singleton().is_started:
+            raise RuntimeError('Cannot register models with the experiment after experiment has started.'
+                               'Register models before calling experiment.start')
+
         self.model_savers.update(models)
 
     def save(self, global_step):
@@ -47,6 +54,11 @@ class CheckpointSaver:
         """
 
         if not self.model_savers:
+            if not self.__no_savers_warned:
+                labml_notice(["No models were registered for saving\n",
+                              "You can register models with ",
+                              ('experiment.add_pytorch_models', Text.value)])
+                self.__no_savers_warned = True
             return
 
         checkpoints_path = pathlib.Path(self.path)
@@ -72,7 +84,12 @@ class CheckpointSaver:
         """
 
         if not self.model_savers:
-            return False
+            if not self.__no_savers_warned:
+                labml_notice(["No models were registered for loading or saving\n",
+                              "You can register models with ",
+                              ('experiment.add_pytorch_models', Text.value)])
+                self.__no_savers_warned = True
+            return
 
         if not models:
             models = list(self.model_savers.keys())
@@ -80,12 +97,34 @@ class CheckpointSaver:
         with open(str(checkpoint_path / "info.json"), "r") as f:
             info = json.loads(f.readline())
 
-        # Load each model
+        to_load = []
+        not_loaded = []
+        missing = []
         for name in models:
+            if name not in info:
+                missing.append(name)
+            else:
+                to_load.append(name)
+        for name in info:
+            if name not in models:
+                not_loaded.append(name)
+
+        # Load each model
+        for name in to_load:
             saver = self.model_savers[name]
             saver.load(checkpoint_path, info[name])
 
-        return True
+        if missing:
+            labml_notice([(f'{missing} ', Text.highlight),
+                          ('model(s) could not be found.\n'),
+                          (f'{to_load} ', Text.none),
+                          ('models were loaded.', Text.none)
+                          ], is_danger=True)
+        if not_loaded:
+            labml_notice([(f'{not_loaded} ', Text.none),
+                          ('models were not loaded.\n', Text.none),
+                          'Models to be loaded should be specified with: ',
+                          ('experiment.add_pytorch_models', Text.value)])
 
 
 class Experiment:
@@ -103,6 +142,8 @@ class Experiment:
             automatically determining ``python_file``
         tags (Set[str], optional): Set of tags for experiment
     """
+    web_api: Optional['WebApiWriter']
+    is_started: bool
     run: Run
     configs_processor: Optional[ConfigProcessor]
 
@@ -110,13 +151,18 @@ class Experiment:
     check_repo_dirty: bool
     checkpoint_saver: CheckpointSaver
 
+    distributed_rank: int
+    distributed_world_size: int
+
     def __init__(self, *,
+                 uuid: str,
                  name: Optional[str],
                  python_file: Optional[str],
                  comment: Optional[str],
                  writers: Set[str],
                  ignore_callers: Set[str],
-                 tags: Optional[Set[str]]):
+                 tags: Optional[Set[str]],
+                 is_evaluate: bool):
         if python_file is None:
             python_file = get_caller_file(ignore_callers)
 
@@ -140,24 +186,21 @@ class Experiment:
         if global_params_singleton().comment is not None:
             comment = global_params_singleton().comment
 
-        self.name = name
         self.experiment_path = lab_singleton().experiments / name
 
         self.check_repo_dirty = lab_singleton().check_repo_dirty
 
         self.configs_processor = None
 
-        experiment_path = pathlib.Path(self.experiment_path)
-        if not experiment_path.exists():
-            experiment_path.mkdir(parents=True)
-
         if tags is None:
             tags = set(name.split('_'))
 
         self.run = Run.create(
+            uuid=uuid,
             experiment_path=self.experiment_path,
             python_file=python_file,
             trial_time=time.localtime(),
+            name=name,
             comment=comment,
             tags=list(tags))
 
@@ -170,30 +213,20 @@ class Experiment:
             self.run.diff = repo.git.diff()
         except git.InvalidGitRepositoryError:
             if not is_colab() and not is_kaggle():
-                warnings.warn(f"Not a valid git repository",
-                              UserWarning, stacklevel=4)
+                labml_notice(["Not a valid git repository: ",
+                              (lab_singleton().path, Text.value)])
             self.run.commit = 'unknown'
             self.run.commit_message = ''
             self.run.is_dirty = True
             self.run.diff = ''
 
-        logger_internal().reset_writers()
-
-        if 'sqlite' in writers:
-            from labml.internal.logger.writers import sqlite
-            logger_internal().add_writer(
-                sqlite.Writer(self.run.sqlite_path, self.run.artifacts_folder))
-        if 'tensorboard' in writers:
-            from labml.internal.logger.writers import tensorboard
-            logger_internal().add_writer(tensorboard.Writer(self.run.tensorboard_log_path))
-        if 'web_api' in writers:
-            from labml.internal.logger.writers import web_api
-            self.web_api = web_api.Writer()
-            logger_internal().add_writer(self.web_api)
-        else:
-            self.web_api = None
-
         self.checkpoint_saver = CheckpointSaver(self.run.checkpoint_path)
+        self.is_evaluate = is_evaluate
+        self.web_api = None
+        self.writers = writers
+        self.is_started = False
+        self.distributed_rank = 0
+        self.distributed_world_size = -1
 
     def __print_info(self):
         """
@@ -202,7 +235,7 @@ class Experiment:
 
         logger.log()
         logger.log([
-            (self.name, Text.title),
+            (self.run.name, Text.title),
             ': ',
             (str(self.run.uuid), Text.meta)
         ])
@@ -227,33 +260,29 @@ class Experiment:
             ])
 
     def _load_checkpoint(self, checkpoint_path: pathlib.Path):
-        if not self.checkpoint_saver.load(checkpoint_path):
-            logger.log('No models registered', Text.warning)
+        self.checkpoint_saver.load(checkpoint_path)
 
     def save_checkpoint(self):
-        self.checkpoint_saver.save(logger_internal().global_step)
+        if self.is_evaluate:
+            return
+        if self.distributed_rank != 0:
+            return
+
+        self.checkpoint_saver.save(tracker().global_step)
 
     def calc_configs(self,
-                     configs: Configs,
-                     configs_override: Optional[Dict[str, any]],
-                     run_order: Optional[List[Union[List[str], str]]]):
+                     configs: Union[Configs, Dict[str, any]],
+                     configs_override: Optional[Dict[str, any]]):
         if configs_override is None:
             configs_override = {}
         if global_params_singleton().configs is not None:
             configs_override.update(global_params_singleton().configs)
 
         self.configs_processor = ConfigProcessor(configs, configs_override)
-        self.configs_processor(run_order)
 
-        logger.log()
+        if self.distributed_rank == 0:
+            logger.log()
 
-    def calc_configs_dict(self,
-                          configs: Dict[str, any],
-                          configs_override: Optional[Dict[str, any]]):
-        self.configs_processor = ConfigProcessorDict(configs, configs_override)
-        self.configs_processor()
-
-        logger.log()
 
     def __start_from_checkpoint(self, run_uuid: str, checkpoint: Optional[int]):
         checkpoint_path, global_step = experiment_run.get_run_checkpoint(
@@ -275,17 +304,66 @@ class Experiment:
                     checkpoint: Optional[int] = None):
         if checkpoint is None:
             checkpoint = -1
-        checkpoint_path, global_step = experiment_run.get_run_checkpoint(
-            run_uuid,
-            checkpoint)
+        checkpoint_path, global_step = experiment_run.get_run_checkpoint(run_uuid, checkpoint)
 
         if global_step is None:
-            warnings.warn(f"Could not find checkpoint",
-                          UserWarning, stacklevel=4)
+            labml_notice(['Could not find saved checkpoint'], is_danger=True)
             return
 
         with monit.section("Loading checkpoint"):
-            _ = self.checkpoint_saver.load(checkpoint_path, models)
+            self.checkpoint_saver.load(checkpoint_path, models)
+
+    def _save_pid(self):
+        if not self.run.pids_path.exists():
+            self.run.pids_path.mkdir()
+
+        pid_path = self.run.pids_path / f'{self.distributed_rank}.pid'
+        assert not pid_path.exists()
+
+        with open(str(pid_path), 'w') as f:
+            f.write(f'{os.getpid()}')
+
+    def distributed(self, rank: int, world_size: int):
+        self.distributed_rank = rank
+        self.distributed_world_size = world_size
+
+        if self.distributed_rank != 0:
+            monitor().silent()
+
+        # to make sure we have the path to save pid
+        self.run.make_path()
+
+    def _start_tracker(self):
+        tracker().reset_writers()
+
+        if self.is_evaluate:
+            return
+
+        if self.distributed_rank != 0:
+            return
+
+        if 'screen' in self.writers:
+            from labml.internal.tracker.writers import screen
+            tracker().add_writer(screen.ScreenWriter())
+
+        if 'sqlite' in self.writers:
+            from labml.internal.tracker.writers import sqlite
+            tracker().add_writer(sqlite.Writer(self.run.sqlite_path, self.run.artifacts_folder))
+
+        if 'tensorboard' in self.writers:
+            from labml.internal.tracker.writers import tensorboard
+            tracker().add_writer(tensorboard.Writer(self.run.tensorboard_log_path))
+
+        if 'file' in self.writers:
+            from labml.internal.tracker.writers import file
+            tracker().add_writer(file.Writer(self.run.log_file))
+
+        if 'web_api' in self.writers:
+            from labml.internal.tracker.writers import web_api
+            self.web_api = web_api.Writer()
+            tracker().add_writer(self.web_api)
+        else:
+            self.web_api = None
 
     def start(self, *,
               run_uuid: Optional[str] = None,
@@ -298,46 +376,56 @@ class Experiment:
             global_step = 0
 
         self.run.start_step = global_step
-        logger_internal().set_start_global_step(global_step)
 
-        self.__print_info()
-        if self.check_repo_dirty and self.run.is_dirty:
-            logger.log([("[FAIL]", Text.danger),
-                        " Cannot trial an experiment with uncommitted changes."])
-            exit(1)
+        self._start_tracker()
+        tracker().set_start_global_step(global_step)
 
-        if self.configs_processor is not None:
-            self.configs_processor.print()
+        if self.distributed_rank == 0:
+            self.__print_info()
+            if self.check_repo_dirty and self.run.is_dirty:
+                logger.log([("[FAIL]", Text.danger),
+                            " Cannot trial an experiment with uncommitted changes."])
+                exit(1)
 
-        self.run.save_info()
+        if not self.is_evaluate:
+            if self.distributed_rank == 0:
+                self.run.save_info()
+            self._save_pid()
 
-        if self.configs_processor is not None:
-            self.configs_processor.save(self.run.configs_path)
+            if self.distributed_rank == 0:
+                if self.configs_processor is not None:
+                    self.configs_processor.add_saver(FileConfigsSaver(self.run.configs_path))
 
-        if self.web_api is not None:
-            self.web_api.set_info(run_uuid=self.run.uuid,
-                                  name=self.name,
-                                  comment=self.run.comment)
-            if self.configs_processor is not None:
-                self.web_api.set_configs(self.configs_processor.to_json())
-            self.web_api.start()
+                if self.web_api is not None:
+                    self.web_api.set_info(run_uuid=self.run.uuid,
+                                          name=self.run.name,
+                                          comment=self.run.comment)
+                    self.web_api.start()
+                    if self.configs_processor is not None:
+                        self.configs_processor.add_saver(self.web_api.get_configs_saver())
 
-        logger_internal().save_indicators(self.run.indicators_path)
-        if self.configs_processor:
-            logger_internal().write_h_parameters(self.configs_processor.get_hyperparams())
+                tracker().save_indicators(self.run.indicators_path)
 
+                # PERF: Writing to tensorboard takes about 4 seconds
+                # Also wont work when configs are updated live
+                # if self.configs_processor:
+                #     tracker().write_h_parameters(self.configs_processor.get_hyperparams())
+
+        self.is_started = True
         return ExperimentWatcher(self)
 
     def finish(self, status: str, details: any = None):
-        with open(str(self.run.run_log_path), 'a') as f:
-            end_time = time.time()
-            data = json.dumps({'status': status,
-                               'details': details,
-                               'time': end_time}, indent=None)
-            f.write(data + '\n')
+        if not self.is_evaluate:
+            with open(str(self.run.run_log_path), 'a') as f:
+                end_time = time.time()
+                data = json.dumps({'status': status,
+                                   'rank': self.distributed_rank,
+                                   'details': details,
+                                   'time': end_time}, indent=None)
+                f.write(data + '\n')
 
         if self.web_api is not None:
-            self.web_api.status(status, details, end_time)
+            self.web_api.status(self.distributed_rank, status, details, end_time)
 
 
 class GlobalParams:
@@ -368,17 +456,21 @@ def experiment_singleton() -> Experiment:
 
 
 def create_experiment(*,
+                      uuid: str,
                       name: Optional[str],
                       python_file: Optional[str],
                       comment: Optional[str],
                       writers: Set[str],
                       ignore_callers: Set[str],
-                      tags: Optional[Set[str]]):
+                      tags: Optional[Set[str]],
+                      is_evaluate: bool):
     global _internal
 
-    _internal = Experiment(name=name,
+    _internal = Experiment(uuid=uuid,
+                           name=name,
                            python_file=python_file,
                            comment=comment,
                            writers=writers,
                            ignore_callers=ignore_callers,
-                           tags=tags)
+                           tags=tags,
+                           is_evaluate=is_evaluate)
